<?php

declare(strict_types=1);

namespace Drupal\smart_migrate_cli\Commands;

use Consolidation\OutputFormatters\StructuredData\RowsOfFields;
use Drupal\Core\Datetime\DateFormatter;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\KeyValueStore\KeyValueFactoryInterface;
use Drupal\Core\KeyValueStore\KeyValueStoreInterface;
use Drupal\Core\Session\AccountSwitcherInterface;
use Drupal\Core\Session\UserSession;
use Drupal\migrate\MigrateMessageInterface;
use Drupal\migrate\Plugin\MigrationInterface;
use Drupal\migrate\Plugin\MigrationPluginManagerInterface;
use Drupal\migrate_drupal\Plugin\MigrationWithFollowUpInterface;
use Drupal\smart_migrate_cli\Traits\MigrateRunnerTrait;
use Drupal\smart_migrate_cli\Traits\SmartMigrationConfigurationTrait;
use Drush\Commands\DrushCommands;
use Drush\Drupal\Migrate\MigrateExecutable;
use Drush\Drupal\Migrate\MigrateMessage;
use Drush\Drupal\Migrate\MigrateUtils;
use Drush\Utils\StringUtils;
use Symfony\Component\Console\Output\OutputInterface;
use Symfony\Component\Filesystem\Path;

/**
 * Smarter Migrate runner commands.
 *
 * Now:
 * - Build optimal execution order on import operation, making
 *   "--execute-dependencies" unnecessary in most of the situations.
 * - Removes unnecessary D6 migrations if we are migrating from Drupal 7 and
 *   vice versa.
 * - Skips the unnecessary node migrations.
 */
class MigrateRunnerCommands extends DrushCommands {

  use SmartMigrationConfigurationTrait;
  use MigrateRunnerTrait;

  /**
   * Storage for warnings happened during import.
   *
   * @var string[]
   */
  protected array $importWarnings;

  /**
   * Storage for follow-up migrations to execute.
   *
   * @var array
   */
  protected array $followUpMigrationsToExecute;

  /**
   * Storage of the last import timestamps.
   *
   * @var \Drupal\Core\KeyValueStore\KeyValueStoreInterface
   */
  protected KeyValueStoreInterface $keyValue;

  /**
   * Cache of the migrate message service.
   *
   * @var \Drupal\migrate\MigrateMessageInterface
   */
  protected MigrateMessageInterface $migrateMessage;

  public function __construct(
    protected readonly DateFormatter $dateFormatter,
    protected readonly KeyValueFactoryInterface $keyValueFactory,
    protected readonly AccountSwitcherInterface $accountSwitcher,
    protected readonly EntityTypeManagerInterface $entityTypeManager,
    protected readonly ModuleHandlerInterface $moduleHandler,
    protected ?MigrationPluginManagerInterface $migrationPluginManager = NULL,
  ) {
    parent::__construct();
    $this->keyValue = $keyValueFactory->get('migrate_last_imported');
  }

  /**
   * Overrides migrate:status command.
   *
   * The body is the exact copy of MigrateRunnerCommands::status().
   *
   * @hook replace-command migrate:status
   *
   * @see \Drush\Commands\core\MigrateRunnerCommands::status()
   */
  public function status(
    ?string $migrationIds = NULL,
    array $options = [
      'tag' => self::REQ,
      'format' => 'table',
    ],
  ): RowsOfFields {
    $fields = [];
    if ($options['field']) {
      $fields = [$options['field']];
    }
    elseif ($options['fields']) {
      $fields = StringUtils::csvToArray($options['fields']);
    }

    $list = $this->getMigrationList($migrationIds, $options['tag']);

    $table = [];
    // Take it one tag at a time, listing the migrations within each tag.
    foreach ($list as $tag => $migrations) {
      if ($tag) {
        $table[] = $this->padTableRow([
          'id' => dt('Tag: @name', ['@name' => $tag]),
        ], $fields);
      }
      ksort($migrations);
      foreach ($migrations as $migration) {
        $row = [];
        foreach ($fields as $field) {
          switch ($field) {
            case 'id':
              $row[$field] = ($tag ? ' ' : '') . $migration->id();
              break;

            case 'status':
              $row[$field] = $migration->getStatusLabel();
              break;

            case 'total':
              $sourceRowsCount = $this->getMigrationSourceRowsCount($migration);
              $row[$field] = $sourceRowsCount ?? dt('N/A');
              break;

            case 'needing_update':
              $row[$field] = $this->getMigrationNeedingUpdateCount($migration);
              break;

            case 'unprocessed':
              $unprocessedCount = $this->getMigrationUnprocessedCount($migration);
              $row[$field] = $unprocessedCount ?? dt('N/A');
              break;

            case 'imported':
              $importedCount = $this->getMigrationImportedCount($migration);
              if ($importedCount === NULL) {
                // Next migration.
                continue 2;
              }
              $sourceRowsCount = $sourceRowsCount ?? $this->getMigrationSourceRowsCount($migration);
              if ($sourceRowsCount > 0 && $importedCount > 0) {
                $importedCount .= ' (' . round(($importedCount / $sourceRowsCount) * 100, 1) . '%)';
              }
              $row[$field] = $importedCount;
              break;

            case 'last_imported':
              $row[$field] = $this->getMigrationLastImportedTime($migration);
              break;
          }
        }
        $table[] = $row;
      }

      // Add an empty row after a tag group.
      if ($tag) {
        $table[] = $this->padTableRow([], $fields);
      }
    }

    return new RowsOfFields($table);
  }

  /**
   * Returns the migration source rows count.
   *
   * @param \Drupal\migrate\Plugin\MigrationInterface $migration
   *   The migration plugin instance.
   *
   * @return int|null
   *   The migration source rows count or null if the source is uncountable or
   *   the source count couldn't be retrieved.
   *
   * @see \Drush\Commands\core\MigrateRunnerCommands::getMigrationSourceRowsCount()
   */
  protected function getMigrationSourceRowsCount(MigrationInterface $migration): ?int {
    try {
      $sourceRowsCount = $migration->getSourcePlugin()->count();
      // -1 indicates uncountable sources.
      if ($sourceRowsCount === -1) {
        return NULL;
      }
      return $sourceRowsCount;
    }
    catch (\Exception $exception) {
      $arguments = [
        '@migration' => $migration->id(),
        '@message' => $exception->getMessage(),
      ];
      $this->logger()->error(dt('Could not retrieve source count from @migration: @message', $arguments));
      return NULL;
    }
  }

  /**
   * Returns the number of items that needs update.
   *
   * @param \Drupal\migrate\Plugin\MigrationInterface $migration
   *   The migration plugin instance.
   *
   * @return int
   *   The number of items that needs update.
   *
   * @see \Drush\Commands\core\MigrateRunnerCommands::getMigrationNeedingUpdateCount()
   */
  protected function getMigrationNeedingUpdateCount(MigrationInterface $migration): int {
    $map = $migration->getIdMap();
    return count($map->getRowsNeedingUpdate($map->processedCount()));
  }

  /**
   * Returns the number of unprocessed items.
   *
   * @param \Drupal\migrate\Plugin\MigrationInterface $migration
   *   The migration plugin instance.
   *
   * @return int|null
   *   The number of unprocessed items or null if it cannot be determined.
   *
   * @see \Drush\Commands\core\MigrateRunnerCommands::getMigrationUnprocessedCount()
   */
  protected function getMigrationUnprocessedCount(MigrationInterface $migration): ?int {
    $sourceRowsCount = $this->getMigrationSourceRowsCount($migration);
    if ($sourceRowsCount === NULL) {
      return NULL;
    }
    return $sourceRowsCount - $migration->getIdMap()->processedCount();
  }

  /**
   * Returns the number of imported items.
   *
   * @param \Drupal\migrate\Plugin\MigrationInterface $migration
   *   The migration plugin instance.
   *
   * @return int|null
   *   The number of imported items or null if it cannot be determined.
   *
   * @see \Drush\Commands\core\MigrateRunnerCommands::getMigrationImportedCount()
   */
  protected function getMigrationImportedCount(MigrationInterface $migration): ?int {
    try {
      return $migration->getIdMap()->importedCount();
    }
    catch (\Exception $exception) {
      $arguments = [
        '@migration' => $migration->id(),
        '@message' => $exception->getMessage(),
      ];
      $this->logger()->error(dt('Failure retrieving information on @migration: @message', $arguments));
      return NULL;
    }
  }

  /**
   * Returns the last imported date/time if any.
   *
   * @param \Drupal\migrate\Plugin\MigrationInterface $migration
   *   The migration plugin instance.
   *
   * @return string
   *   The last imported date/time if any.
   *
   * @see \Drush\Commands\core\MigrateRunnerCommands::getMigrationLastImportedTime()
   */
  protected function getMigrationLastImportedTime(MigrationInterface $migration): string {
    if ($lastImported = $this->keyValue->get($migration->id(), '')) {
      $lastImported = $this->dateFormatter->format(round($lastImported / 1000), 'custom', 'Y-m-d H:i:s');
    }
    return $lastImported;
  }

  /**
   * Pads an incomplete table row with empty cells.
   *
   * @param array $row
   *   The row to be prepared.
   * @param array $fields
   *   The table columns.
   *
   * @return array
   *   The complete table row.
   *
   * @see \Drush\Commands\core\MigrateRunnerCommands::padTableRow()
   */
  protected function padTableRow(array $row, array $fields): array {
    foreach (array_diff_key(array_flip($fields), $row) as $field => $delta) {
      $row[$field] = NULL;
    }
    return $row;
  }

  /**
   * Override for migrate:rollback command.
   *
   * @hook replace-command migrate:rollback
   *
   * @see \Drush\Commands\core\MigrateRunnerCommands::rollback()
   */
  public function rollback(
    ?string $migrationIds = NULL,
    array $options = [
      'all' => FALSE,
      'tag' => self::REQ,
      'feedback' => self::REQ,
      'idlist' => self::REQ,
      'progress' => TRUE,
    ],
  ): void {
    $tags = $options['tag'];
    $all = $options['all'];

    if (!$all && !$migrationIds && !$tags) {
      throw new \Exception(dt('You must specify --all, --tag, or one or more migration names separated by commas'));
    }

    if (!$list = $this->getMigrationList($migrationIds, $options['tag'])) {
      $this->logger()->error(dt('No migrations found.'));
    }

    $executableOptions = array_intersect_key(
      $options,
      array_flip(['feedback', 'idlist', 'progress'])
    );
    foreach ($list as $migrations) {
      // Rollback in reverse order.
      $migrations = array_reverse($migrations);
      foreach ($migrations as $migration) {
        $executable = new MigrateExecutable($migration, $this->getMigrateMessage(), $this->output(), $executableOptions);
        // drush_op() provides --simulate support.
        drush_op([$executable, 'rollback']);
      }
    }
  }

  /**
   * Retrieves a list of active migrations.
   *
   * @param string|null $migrationIds
   *   A comma-separated list of migration IDs. If omitted, will return all
   *   migrations.
   * @param string|null $tags
   *   A comma separated list of tags.
   *
   * @return \Drupal\migrate\Plugin\MigrationInterface[][]
   *   An array keyed by migration tag, each value containing an array of
   *   migrations or an empty array if no migrations match the input criteria.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   *
   * @see \Drush\Commands\core\MigrateRunnerCommands::getMigrationList()
   */
  protected function getMigrationList(?string $migrationIds, ?string $tags): array {
    $invoked = debug_backtrace(DEBUG_BACKTRACE_IGNORE_ARGS, 2)[1]['function'] ?? NULL;
    $order_forced = func_get_args()[2] ?? FALSE;
    $is_import_operation = $invoked === 'import' || $order_forced;
    $grouped_list = $this->doGetGroupedMigrationList($migrationIds, $tags);
    if ($is_import_operation) {
      ksort($grouped_list);
    }

    $grouped_list = $this->filterMigrationList($grouped_list);

    return $this->orderMigrationList($grouped_list, $is_import_operation);
  }

  /**
   * Executes one or more migrations.
   *
   * @hook replace-command migrate:import
   *
   * @see \Drush\Commands\core\MigrateRunnerCommands::import()
   */
  public function import(
    ?string $migrationIds = NULL,
    array $options = [
      'all' => FALSE,
      'tag' => self::REQ,
      'limit' => self::REQ,
      'feedback' => self::REQ,
      'idlist' => self::REQ,
      'update' => FALSE,
      'force' => FALSE,
      'execute-dependencies' => FALSE,
      'timestamp' => FALSE,
      'total' => FALSE,
      'progress' => TRUE,
      'delete' => FALSE,
    ],
  ): void {
    $tags = $options['tag'];
    $all = $options['all'];

    if (!$all && !$migrationIds && !$tags) {
      throw new \Exception(dt('You must specify --all, --tag or one or more migration names separated by commas'));
    }

    if (!$list = $this->getMigrationList($migrationIds, $options['tag'])) {
      throw new \Exception(dt('No migrations found.'));
    }

    $userData = [
      'options' => array_intersect_key($options, array_flip([
        'limit',
        'feedback',
        'idlist',
        'update',
        'force',
        'timestamp',
        'total',
        'progress',
        'delete',
      ])),
      'execute_dependencies' => $options['execute-dependencies'],
    ];

    $drushBase = defined('DRUSH_BASE_PATH')
      ? DRUSH_BASE_PATH
      : $this->getConfig()->get('drush.base-dir');

    // Include the file providing a migrate_prepare_row hook implementation.
    require_once Path::join($drushBase, 'src/Drupal/Migrate/migrate_runner.inc');
    // If the 'migrate_prepare_row' hook implementations are already cached,
    // make sure that system_migrate_prepare_row() is picked-up.
    $this->moduleHandler->resetImplementations();

    // Switch to user 1. This is what we usually do on the UI as well.
    // The user ID mostly affects the user reference (base) fields of migrated
    // entities which don't have source data - e.g. the owner of user or
    // taxonomy term translations.
    $this->accountSwitcher->switchTo(new UserSession(['uid' => 1]));
    $this->importWarnings = [];
    $this->followUpMigrationsToExecute = [];
    foreach ($list as $migrations) {
      array_walk($migrations, [static::class, 'executeMigration'], $userData);
    }
    if (!empty($this->followUpMigrationsToExecute)) {
      $this->writeLineIfNotQuiet('');
      $this->logger()->notice('Executing follow-up migrations');
      $this->writeLineIfNotQuiet('');
      array_walk(
        $this->followUpMigrationsToExecute,
        [static::class, 'executeMigration'],
        $userData
      );
    }
    $this->accountSwitcher->switchBack();
    if ($this->importWarnings) {
      $this->writeLineIfNotQuiet('');
      $lines_to_log = array_merge(
        [dt('Some rows failed to import, but this not necessarily means that something went terribly wrong. Consider checking the migration map and migration message tables of the following migrations.')],
        $this->importWarnings
      );
      $this->logger()->warning(implode(
        "\n           ",
        $lines_to_log
      ));
    }
  }

  /**
   * Executes a single migration.
   *
   * If the --execute-dependencies option was given, the migration's
   * dependencies will also be executed first.
   *
   * @param \Drupal\migrate\Plugin\MigrationInterface $migration
   *   The migration to execute.
   * @param string $migrationId
   *   The migration ID (not used, just an artifact of array_walk()).
   * @param array $userData
   *   Additional data passed to the callback.
   *
   * @throws \Exception
   *   If there are failed migrations.
   *
   * @see \Drush\Commands\core\MigrateRunnerCommands::executeMigration()
   */
  protected function executeMigration(MigrationInterface $migration, string $migrationId, array $userData): void {
    $migration = $this->getFreshMigrationInstance($migration->id());

    $this->prepareMigration($migration);

    try {
      $this->originalExecuteMigration($migration, $migrationId, $userData);
    }
    catch (\Exception $e) {
      $this->importWarnings[] = $e->getMessage();
    }

    // After the migration on which they depend has been successfully executed,
    // the follow-up migrations should be immediately added to the list of the
    // migrations to be executed.
    if ($migration instanceof MigrationWithFollowUpInterface) {
      $diff = array_diff_key($migration->generateFollowUpMigrations(), $this->followUpMigrationsToExecute ?? []);
      $this->followUpMigrationsToExecute = array_merge(
        $this->followUpMigrationsToExecute ?? [],
        $diff
      );
    }
  }

  /**
   * Writes a message to non-quiet output and adds a newline at the end.
   *
   * @param string|iterable $messages
   *   The message.
   * @param int $options
   *   A bitmask of options (one of the OUTPUT or VERBOSITY constants), 0 is
   *   considered the same as self::OUTPUT_NORMAL | self::VERBOSITY_NORMAL.
   */
  protected function writeLineIfNotQuiet($messages, int $options = 0) {
    if ($this->output()->getVerbosity() !== OutputInterface::VERBOSITY_QUIET) {
      $this->output()->writeln($messages, $options);
    }
  }

  /**
   * Executes a single migration.
   *
   * If the --execute-dependencies option was given, the migration's
   * dependencies will also be executed first.
   *
   * @param \Drupal\migrate\Plugin\MigrationInterface $migration
   *   The migration to execute.
   * @param string $migrationId
   *   The migration ID (not used, just an artifact of array_walk()).
   * @param array $userData
   *   Additional data passed to the callback.
   *
   * @throws \Exception
   *   If there are failed migrations.
   *
   * @see \Drush\Commands\core\MigrateRunnerCommands::executeMigration()
   */
  protected function originalExecuteMigration(MigrationInterface $migration, string $migrationId, array $userData): void {
    static $executedMigrations = [];

    if ($userData['execute_dependencies']) {
      $dependencies = $migration->getMigrationDependencies()['required'];
      // Remove already executed migrations.
      $dependencies = array_diff($dependencies, $executedMigrations);
      if ($dependencies) {
        $requiredMigrations = $this->migrationPluginManager->createInstances($dependencies);
        array_walk($requiredMigrations, [static::class, __FUNCTION__], $userData);
      }
    }
    if (!empty($userData['options']['force'])) {
      // @todo Use the new MigrationInterface::setRequirements() method,
      //   instead of Migration::set() and remove the PHPStan exception
      //   from phpstan-baseline.neon when #2796755 lands in Drupal core.
      // @see https://www.drupal.org/i/2796755
      $migration->set('requirements', []);
    }
    if (!empty($userData['options']['update'])) {
      if (empty($userData['options']['idlist'])) {
        $migration->getIdMap()->prepareUpdate();
      }
      else {
        $sourceIdValuesList = MigrateUtils::parseIdList($userData['options']['idlist']);
        $keys = array_keys($migration->getSourcePlugin()->getIds());
        foreach ($sourceIdValuesList as $sourceIdValues) {
          $migration->getIdMap()->setUpdate(array_combine($keys, $sourceIdValues));
        }
      }
    }

    $executable = new MigrateExecutable($migration, $this->getMigrateMessage(), $this->output(), $userData['options']);
    // drush_op() provides --simulate support.
    drush_op([$executable, 'import']);
    if ($count = $executable->getFailedCount()) {
      // Nudge Drush to use a non-zero exit code.
      throw new \Exception(dt('!name migration: !count failed.', ['!name' => $migrationId, '!count' => $count]));
    }

    // Keep track of executed migrations.
    $executedMigrations[] = $migrationId;
  }

  /**
   * Returns the migrate message logger.
   *
   * @return \Drupal\migrate\MigrateMessageInterface
   *   The migrate message logger.
   *
   * @see \Drush\Commands\core\MigrateRunnerCommands::getMigrateMessage()
   */
  protected function getMigrateMessage(): MigrateMessageInterface {
    if (!isset($this->migrateMessage)) {
      $this->migrateMessage = new MigrateMessage($this->logger());
    }
    return $this->migrateMessage;
  }

}
